<?php
/* vim: set ts=4 sw=4 sts=4 et: */
/*****************************************************************************\
 * +-----------------------------------------------------------------------------+
 * | X-Cart Software license agreement                                           |
 * | Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>         |
 * | All rights reserved.                                                        |
 * +-----------------------------------------------------------------------------+
 * | PLEASE READ  THE FULL TEXT OF SOFTWARE LICENSE AGREEMENT IN THE "COPYRIGHT" |
 * | FILE PROVIDED WITH THIS DISTRIBUTION. THE AGREEMENT TEXT IS ALSO AVAILABLE  |
 * | AT THE FOLLOWING URL: https://www.x-cart.com/license-agreement-classic.html |
 * |                                                                             |
 * | THIS AGREEMENT EXPRESSES THE TERMS AND CONDITIONS ON WHICH YOU MAY USE THIS |
 * | SOFTWARE PROGRAM AND ASSOCIATED DOCUMENTATION THAT QUALITEAM SOFTWARE LTD   |
 * | (hereinafter referred to as "THE AUTHOR") OF REPUBLIC OF CYPRUS IS          |
 * | FURNISHING OR MAKING AVAILABLE TO YOU WITH THIS AGREEMENT (COLLECTIVELY,    |
 * | THE "SOFTWARE"). PLEASE REVIEW THE FOLLOWING TERMS AND CONDITIONS OF THIS   |
 * | LICENSE AGREEMENT CAREFULLY BEFORE INSTALLING OR USING THE SOFTWARE. BY     |
 * | INSTALLING, COPYING OR OTHERWISE USING THE SOFTWARE, YOU AND YOUR COMPANY   |
 * | (COLLECTIVELY, "YOU") ARE ACCEPTING AND AGREEING TO THE TERMS OF THIS       |
 * | LICENSE AGREEMENT. IF YOU ARE NOT WILLING TO BE BOUND BY THIS AGREEMENT, DO |
 * | NOT INSTALL OR USE THE SOFTWARE. VARIOUS COPYRIGHTS AND OTHER INTELLECTUAL  |
 * | PROPERTY RIGHTS PROTECT THE SOFTWARE. THIS AGREEMENT IS A LICENSE AGREEMENT |
 * | THAT GIVES YOU LIMITED RIGHTS TO USE THE SOFTWARE AND NOT AN AGREEMENT FOR  |
 * | SALE OR FOR TRANSFER OF TITLE. THE AUTHOR RETAINS ALL RIGHTS NOT EXPRESSLY  |
 * | GRANTED BY THIS AGREEMENT.                                                  |
 * +-----------------------------------------------------------------------------+
 * \*****************************************************************************/

/**
 * PayPal Checkout
 *
 * @category   X-Cart
 * @package    X-Cart
 * @subpackage Modules
 * @author     Ildar Amankulov
 * @copyright  Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>
 * @license    https://www.x-cart.com/license-agreement-classic.html X-Cart license agreement
 * @version    $Id$
 * @link       https://www.x-cart.com/
 * @see        ____file_see____
 */

if (!defined('XCART_START')) {
    header('Location: ../../');
    die('Access denied');
}

function func_cc_paypal_checkout_get_refund_mode($payment_id, $order_id)
{
    return 'P';
}

function func_cc_paypal_checkout_do_void($order)
{
    return func_cc_paypal_checkout_do($order, 'V');
}

function func_cc_paypal_checkout_do_capture($order)
{
    return func_cc_paypal_checkout_do($order, 'C');
}

function func_cc_paypal_checkout_do_refund($order, $amount = false)
{
    return func_cc_paypal_checkout_do($order, 'R', $amount);
}

function func_cc_paypal_checkout_do($order, $mode, $amount = false)
{
    $error_message = '';
    $status        = true;
    try {
        switch ($mode) {
            case 'C':
                $status = func_paypal_checkout_do_capture($order['order']);
                break;
            case 'V':
                $status = func_paypal_checkout_do_void($order['order']);
                break;
            case 'R':
                $status = func_paypal_checkout_do_refund($order['order'], $amount);
                break;
        }
    } catch (Exception $e) {
        $error_message = $e->getMessage();
        $status        = false;
    }

    return [$status, $error_message, []];
}

function func_pc_handler1_create_order($payment_id, $call_type)
{
    global $PPEC_POST_VARS, $logged_userid, $login, $login_type, $paypal_express_details, $paypal_payment_id, $paypal_token, $paypal_mode, $paypal_token_ttl, $paypal_begin_express;
    global $config, $xcart_catalogs, $smarty, $current_location, $top_message;

    $id = \XC\PaypalCheckout\RestAPI::getInstance()->createOrder();
    if ($call_type !== 'start_express_checkout') {
        return $id;
    }

    x_session_register('paypal_token');
    $pp_return_url = $current_location . '/payment/ps_paypal_pro.php?mode=express_return';
    if (!empty($useraction) && $useraction == 'commit') {
        $pp_return_url .= '&amp;useraction=commit';
        $pp_cancel_url = $xcart_catalogs['customer'] . '/cart.php?mode=checkout&amp;express_cancel';
    } else {
        $pp_cancel_url = $xcart_catalogs['customer'] . '/cart.php?express_cancel';
    }

    x_session_register('paypal_begin_express');
    $paypal_begin_express = false;

    if (defined('XC_SESSION_DB_SAVE_PAYPAL_EXPRESS')) {
        x_session_save('paypal_begin_express');
    }

    x_session_register('paypal_payment_id');
    x_session_register('paypal_mode');

    $paypal_payment_id = $payment_id;
    $paypal_mode       = 'express';

    if (!empty($pp_subject)) {
        $pp_username      = '';
        $pp_password      = '';
        $pp_use_cert      = false;
        $pp_signature_txt = '';
        $pp_final_action  = 'Sale';
    }

    if ($id) {
        $paypal_token = $id;
        x_session_register('paypal_token_ttl');
        $paypal_token_ttl = XC_TIME;
    } else {
        $paypal_token = '';
    }

    return $id;
}

/*
 * Handler when commit=false and a customer passed 1st step on PayPal side, userdata etc
 */
function func_pc_handler2_update_xc_customer_data($paypal_token, $pp_subject, $useraction, $RefererCookie)
{
    global $PPEC_POST_VARS, $logged_userid, $login, $login_type, $paypal_express_details, $paypal_payment_id;
    global $config, $xcart_catalogs, $smarty, $current_location, $top_message;

    x_load('paypal', 'payment');

    $result = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/checkout/orders/{$paypal_token}", [], 'GET');

    if (empty($result['result_data'])) {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg(__FUNCTION__ . ": empty(\$result['result_data']) v2/checkout/orders/{$paypal_token}", true);
    }

    if ($result['status_code'] != 200) {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg(__FUNCTION__ . ": result['status_code'] v2/checkout/orders/{$paypal_token}" . print_r($result, true), true);
    }

    $module_params = func_get_pm_params('ps_paypal_pro.php');

    if ($pp_subject) {
        $pp_username      = '';
        $pp_password      = '';
        $pp_use_cert      = false;
        $pp_signature_txt = '';
        $pp_final_action  = 'Sale';
    }

    $state_err = 0;

    x_session_register('login');
    x_session_register('login_type');
    x_session_register('logged_userid');
    $result = $result['result_data'];

    if (empty($useraction) || $useraction != 'commit') {
        $paypal_s_address = empty($result['purchase_units'][0]['shipping']['address']) ? $result['payer']['address'] : $result['purchase_units'][0]['shipping']['address'];
        list($paypal_s_firstname, $paypal_s_lastname) = array_merge(explode(" ",
            isset($result['purchase_units'][0]['shipping']['name']['full_name']) ? $result['purchase_units'][0]['shipping']['name']['full_name'] : ''
        ), ['']);
        $address = [
            'firstname' => !empty($result['payer']['name']['given_name']) ? $result['payer']['name']['given_name'] : $paypal_s_firstname,
            'lastname'  => !empty($result['payer']['name']['surname']) ? $result['payer']['name']['surname'] : $paypal_s_lastname,
            'address'   => preg_replace('![\s\n\r]+!s', ' ', $paypal_s_address['address_line_1']),
            'city'      => $paypal_s_address['admin_area_2'],
            'state'     => func_paypal_detect_state($paypal_s_address['country_code'], $paypal_s_address['admin_area_1'], $paypal_s_address['postal_code'], $state_err),
            'country'   => $paypal_s_address['country_code'],
            'zipcode'   => $paypal_s_address['postal_code'],
            'phone'     => !empty($result['phone']['phone_number']['national_number']) ? $result['phone']['phone_number']['national_number'] : '',
        ];

        if (!empty($paypal_s_address['address_line_2'])) {
            $address['address'] .= "\n" . preg_replace('![\s\n\r]+!s', ' ', $paypal_s_address['address_line_2']);
        }

        if ($config['General']['use_counties'] == 'Y') {
            $default_county    = func_default_county($address['state'], $address['country']);
            $address['county'] = empty($default_county) ? $paypal_s_address['admin_area_1'] : $default_county;
        }

        $parsed_profile              = [];
        $parsed_profile['firstname'] = $result['payer']['name']['given_name'];
        $parsed_profile['lastname']  = $result['payer']['name']['surname'];
        $parsed_profile['phone']     = (!empty($result['phone']['phone_number']['national_number']) ? $result['phone']['phone_number']['national_number'] : '');
        $parsed_profile['email']     = $result['payer']['email_address'];
        $parsed_profile['referer']   = !empty($RefererCookie) ? $RefererCookie : '';

        func_paypal_save_profile($address, $parsed_profile);
    }

    if ((empty($login) || $login_type != 'C') && $config['General']['enable_anonymous_checkout'] != 'Y') {
        // Display a warning message about expired session
        $top_message = [
            'type'    => 'E',
            'content' => func_get_langvar_by_name('txt_paypal_expired_session_warn'),
        ];
        func_header_location($xcart_catalogs['customer'] . '/login.php');
    }

    x_session_register('paypal_payment_id');
    x_session_register('paypal_express_details');
    $paypal_express_details = [
        'Token'   => $paypal_token,
        'PayerID' => !empty($result['payer']['payer_id']) ? $result['payer']['payer_id'] : '',
    ];

    switch ($state_err) {
        case 1:
            $top_message = [
                'type'    => 'W',
                'content' => func_get_langvar_by_name('lbl_paypal_wrong_country_note'),
            ];
            break;

        case 2:
        case 3:
            $top_message = [
                'type'    => 'W',
                'content' => func_get_langvar_by_name('lbl_paypal_wrong_state_note'),
            ];
    }

    if (!empty($state_err)) {
        // Redirect to Edit Profile page
        func_header_location($xcart_catalogs['customer'] . '/cart.php?mode=checkout&paymentid=' . $paypal_payment_id . '&edit_profile');
    } elseif (!empty($useraction) && $useraction === 'commit') {
        // Skip Order Confirmation page and submit order
        x_session_register('PPEC_POST_VARS');
        $smarty->webmaster_mode = false;
        $smarty->assign('payment_wait_body_js_code', 'onLoad="document.process.submit();"');

        func_flush(func_display('customer/main/payment_wait.tpl', $smarty, false));
        ?>
        <form action="<?php echo $current_location . '/payment/payment_cc.php'; ?>" method="post" name="process">
            <?php foreach ($PPEC_POST_VARS as $k => $v) { ?>
                <input type="hidden" name="<?php echo $k; ?>" value="<?php echo $v; ?>"/>
            <?php } ?>
        </form>
        </body>
        </html>

        <?php
        exit;
    } else {
        // Redirect to Order Confirmation page
        func_header_location($xcart_catalogs['customer'] . '/cart.php?mode=checkout&paymentid=' . $paypal_payment_id);
    }
}

/**
 * Finish order. commit=false (double step), scenario
 */
function func_pc_handler3_update_n_finish_order()
{
    global $pp_subject, $sql_tbl, $order_secureid, $XCARTSESSID, $secure_oid, $pp_prefix, $cart, $pp_total, $paypal_express_details, $bill_output, $extra_order_data;
    global $xcart_dir, $xcart_catalogs, $config;

    x_load('cart', 'paypal');
    x_session_register('paypal_express_details');

    // step1 update order
    $updateResult = !empty($paypal_express_details['Token'])
        ? \XC\PaypalCheckout\RestAPI::getInstance()->updateOrder($paypal_express_details['Token'])
        : [];

    if (empty($updateResult['status_code']) || $updateResult['status_code'] != 204) {
        func_paypal_express_enable_1step();
        func_paypal_clear_ec_token();
        func_header_location($xcart_catalogs['customer'] . '/cart.php');
    }

    $extra_order_data = [];

    // step2 authorize/capture order
    if ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y') {
        $order_details = \XC\PaypalCheckout\RestAPI::getInstance()->authorizeOrder($paypal_express_details['Token'], ['headers' => ['Prefer: return=representation']]);

        $extra_order_data['capture_status'] = 'A';

    } else {
        $order_details = \XC\PaypalCheckout\RestAPI::getInstance()->captureOrder($paypal_express_details['Token'], ['headers' => ['Prefer: return=representation']]);

        $extra_order_data['capture_status'] = '';

        $captured_data = isset($order_details['purchase_units'][0]['payments']['captures'][0]) ? $order_details['purchase_units'][0]['payments']['captures'][0] : [];

        if (!$captured_data) {
            $order_details['status'] = $order_details['status'] == 'COMPLETED' ? 'notCOMPLETED' : $order_details['status'];
            \XC\PaypalCheckout\RestAPI::getInstance()->logMsg(__FUNCTION__ . ": captureOrder(\$order_details" . print_r($order_details, true), true);
        } else {
            $_response_capture = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/captures/{$captured_data['id']}", [], 'GET');
            if ($_response_capture['status_code'] == 200) {
                $extra_order_data['paypal_checkout_capture'] = addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($_response_capture['result_data']));
            }
        }
    }

    $extra_order_data['paypal_checkout_id']    = addslashes($order_details['id']);
    $extra_order_data['paypal_checkout_order'] = addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($order_details));
    if (version_compare($config['version'], '4.5.0', '>=')) {
        db_query("REPLACE INTO $sql_tbl[cc_pp3_data] (`ref`,`sessid`,`trstat`) VALUES ('{$extra_order_data['paypal_checkout_id']}','$XCARTSESSID','GO|" . implode('|', $secure_oid) . "')");
    } else {
        db_query("REPLACE INTO $sql_tbl[cc_pp3_data] (`ref`,`sessionid`,`trstat`) VALUES ('{$extra_order_data['paypal_checkout_id']}','$XCARTSESSID','GO|" . implode('|', $secure_oid) . "')");
    }

    $bill_output = [];

    if ($order_details['status'] == 'COMPLETED') {
        $bill_output['code'] = 1;

        if ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y') {
            $bill_output['is_preauth'] = true;
        }

        $bill_output['billmes'] = "Payment complete (PayPal Order ID: {$order_details['id']}) ";
    } else {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg(__FUNCTION__ . ": payment was failed" . print_r($order_details, true), true);
    }

    $extra_order_data['paypal_type'] = 'USEC';

    func_paypal_express_enable_1step();
    func_paypal_clear_ec_token();
}

