<?php
/* vim: set ts=4 sw=4 sts=4 et: */
/*****************************************************************************\
 * +-----------------------------------------------------------------------------+
 * | X-Cart Software license agreement                                           |
 * | Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>         |
 * | All rights reserved.                                                        |
 * +-----------------------------------------------------------------------------+
 * | PLEASE READ  THE FULL TEXT OF SOFTWARE LICENSE AGREEMENT IN THE "COPYRIGHT" |
 * | FILE PROVIDED WITH THIS DISTRIBUTION. THE AGREEMENT TEXT IS ALSO AVAILABLE  |
 * | AT THE FOLLOWING URL: https://www.x-cart.com/license-agreement-classic.html |
 * |                                                                             |
 * | THIS AGREEMENT EXPRESSES THE TERMS AND CONDITIONS ON WHICH YOU MAY USE THIS |
 * | SOFTWARE PROGRAM AND ASSOCIATED DOCUMENTATION THAT QUALITEAM SOFTWARE LTD   |
 * | (hereinafter referred to as "THE AUTHOR") OF REPUBLIC OF CYPRUS IS          |
 * | FURNISHING OR MAKING AVAILABLE TO YOU WITH THIS AGREEMENT (COLLECTIVELY,    |
 * | THE "SOFTWARE"). PLEASE REVIEW THE FOLLOWING TERMS AND CONDITIONS OF THIS   |
 * | LICENSE AGREEMENT CAREFULLY BEFORE INSTALLING OR USING THE SOFTWARE. BY     |
 * | INSTALLING, COPYING OR OTHERWISE USING THE SOFTWARE, YOU AND YOUR COMPANY   |
 * | (COLLECTIVELY, "YOU") ARE ACCEPTING AND AGREEING TO THE TERMS OF THIS       |
 * | LICENSE AGREEMENT. IF YOU ARE NOT WILLING TO BE BOUND BY THIS AGREEMENT, DO |
 * | NOT INSTALL OR USE THE SOFTWARE. VARIOUS COPYRIGHTS AND OTHER INTELLECTUAL  |
 * | PROPERTY RIGHTS PROTECT THE SOFTWARE. THIS AGREEMENT IS A LICENSE AGREEMENT |
 * | THAT GIVES YOU LIMITED RIGHTS TO USE THE SOFTWARE AND NOT AN AGREEMENT FOR  |
 * | SALE OR FOR TRANSFER OF TITLE. THE AUTHOR RETAINS ALL RIGHTS NOT EXPRESSLY  |
 * | GRANTED BY THIS AGREEMENT.                                                  |
 * +-----------------------------------------------------------------------------+
 * \*****************************************************************************/

/**
 * PayPal Checkout
 *
 * @category   X-Cart
 * @package    X-Cart
 * @subpackage Modules
 * @author     Ildar Amankulov
 * @copyright  Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>
 * @license    https://www.x-cart.com/license-agreement-classic.html X-Cart license agreement
 * @version    $Id$
 * @link       https://www.x-cart.com/
 * @see        ____file_see____
 */

namespace XC\PaypalCheckout;

if (!defined('XCART_START')) {
    header('Location: ../../');
    die('Access denied');
}

class RestAPI extends \XC_Singleton
{
    const INTEGRATION_DATE   = '2023-09-05'; // https://developer.paypal.com/sdk/js/configuration/#link-integrationdate
    const RENEW_ACCESS_TOKEN = true;

    private $respHeaders = [];

    /**
     * @return RestAPI
     */
    public static function getInstance()
    { // {{{
        // Call parent getter
        return parent::getClassInstance(__CLASS__);
    } // }}}

    public function assignJsWidgetParams()
    {
        global $smarty, $PHP_SELF, $mode, $block, $config, $checkout_module;

        $is_flc = !empty($checkout_module) && $checkout_module === 'Fast_Lane_Checkout';

        $isCheckoutPage =
            (stripos(basename($PHP_SELF), 'cart.php') !== false && !empty($mode) && $mode === 'checkout')
            || (stripos(basename($PHP_SELF), 'get_block.php') !== false && !empty($block) && stripos($block, 'opc_') !== false);

        $components = ['buttons', 'funding-eligibility'];

        $enabled = $config['Paypal_Checkout']['paypal_checkout_enable_funding'] !== 'N' && $config['Paypal_Checkout']['paypal_checkout_enable_funding']
            ? explode(';', $config['Paypal_Checkout']['paypal_checkout_enable_funding']) : [];

        if (!$isCheckoutPage) {
            if (in_array('paylater', $enabled) && $config['Paypal_Checkout']['pc_credit_messaging'] === 'Y') {
                $components[] = 'messages';
            }
            $disabled = ['card'];
        } else {
            // checkout page
            $components = array_merge($components, ['hosted-fields']);

            $disabled = $config['Paypal_Checkout']['paypal_checkout_disable_funding'] !== 'N' && $config['Paypal_Checkout']['paypal_checkout_disable_funding']
                ? explode(';', $config['Paypal_Checkout']['paypal_checkout_disable_funding']) : [];

            if (!in_array('venmo', $disabled)) {
                $enabled = array_merge($enabled, ['venmo']);
            }
            if ($is_flc) {
                foreach ($disabled as $k => $pay_type) {
                    if ($pay_type === 'card') {
                        unset($disabled[$k]);
                    }
                }
                $disabled = array_values($disabled);
            }
        }

        if (!in_array('credit', $disabled)) {
            // #XCB-840 force enable pay later to work properly for AU, FR, GB, DE
            $enabled = array_merge($enabled, ['credit', 'paylater']);
        }

        $paypalCheckoutScriptUrl = 'client-id=' . $config['Paypal_Checkout']['paypal_checkout_client_id'] .
            "&merchant-id={$config['Paypal_Checkout']['pc_merchant_id_of_client']}" . // https://developer.paypal.com/sdk/js/configuration/#link-merchantid
            "&currency={$config['Paypal_Checkout']['paypal_checkout_currency']}" .
            '&intent=' . ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y' ? 'authorize' : 'capture') .
            '&components=' . implode(',', array_unique($components));
        if (!empty($enabled)) {
            $paypalCheckoutScriptUrl .= '&enable-funding=' . implode(',', array_unique($enabled));
        }
        if (!empty($disabled)) {
            $paypalCheckoutScriptUrl .= '&disable-funding=' . implode(',', array_unique($disabled));
        }
        $paypalCheckoutScriptUrl .= '&integration-date=' . static::INTEGRATION_DATE;
        //$paypalCheckoutScriptUrl .= '&buyer-country=BE&debug=true'; // debug paypal credit

        $smarty->assign('paypal_checkout_payment_id', $this->getPaymentId());

        $paypalConfig = [
            'is_checkout_page'          => $isCheckoutPage,
            'is_flc'                    => $is_flc,
            'paypal_checkout_3d_secure' => $config['Paypal_Checkout']['paypal_checkout3d_secure'] === 'Y' ? 'SCA_ALWAYS' : 'SCA_WHEN_REQUIRED',
        ];

        $buttonStyle = $this->getButtonStyle();
        if ($config['Paypal_Checkout']['paypal_checkout_mode'] === 'test') {
            $buttonStyle['env'] = 'sandbox';
        }
        $smarty->assign('paypal_checkout_button_style', $buttonStyle);

        $smarty->assign('paypal_checkout_script', $paypalCheckoutScriptUrl);
        $smarty->assign('paypal_checkout_client_token', $this->generateClientToken());
        $smarty->assign('paypal_checkout_config', json_encode($paypalConfig));
        $smarty->assign('is_checkout_page', $isCheckoutPage);
    }

    public function createOrder()
    {
        return $this->apiCall('v2/checkout/orders', ['data' => $this->getOrderDetails()], 'POST')['result_data']['id'];
    }

    public function updateOrder($order_id)
    {
        return $this->apiCall(
            "v2/checkout/orders/{$order_id}",
            ['data' => [[
                            'op'    => 'replace',
                            'path'  => '/purchase_units/@reference_id==\'default\'',
                            'value' => $this->getOrderDetails()['purchase_units'][0],
                        ]]], 'PATCH');
    }

    public function authorizeOrder($order_id, $params = [])
    {
        return $this->apiCall("v2/checkout/orders/{$order_id}/authorize", $params, 'POST')['result_data'];
    }

    public function captureOrder($order_id, $params = [])
    {
        return $this->apiCall("v2/checkout/orders/{$order_id}/capture", $params, 'POST')['result_data'];
    }

    public function captureAuthorizedPayment($order_data, $order_id = '')
    {
        global $config;
        $authorization_id = 0;
        foreach ($order_data['purchase_units'] as $purchase_unit) {
            foreach ($purchase_unit['payments']['authorizations'] as $authorization) {
                $authorization_id = $authorization['id'];
                break;
            }
        }
        if (empty($authorization_id)) {
            return false;
        }

        if (!empty($order_id)) {
            $params['data']['invoice_id'] = "{$config['Paypal_Checkout']['paypal_checkout_order_prefix']}$order_id";
        } else {
            $params = ['data' => []];
        }

        // https://developer.paypal.com/docs/api/payments/v2/#authorizations_capture
        $captured_response = $this->apiCall("v2/payments/authorizations/{$authorization_id}/capture", $params)['result_data'];
        if (empty($captured_response['status']) || $captured_response['status'] != 'COMPLETED') {
            return false;
        }

        return $captured_response;
    }

    public function refundOrder($capture_data, $order_id = '', $amount = false)
    {
        global $config;
        $params = ['data' => []];

        if (!empty($order_id)) {
            $params['data']['invoice_id'] = "{$config['Paypal_Checkout']['paypal_checkout_order_prefix']}$order_id";
        }

        if (!empty($amount)) {
            $params['data']['amount'] = $amount;
        }

        $refunded_response = $this->apiCall("v2/payments/captures/{$capture_data['id']}/refund", $params)['result_data'];

        if (empty($refunded_response['status']) || $refunded_response['status'] != 'COMPLETED') {
            return false;
        }

        return $refunded_response;
    }

    public function voidOrder($order_data)
    {
        $authorization_id = 0;
        foreach ($order_data['purchase_units'] as $purchase_unit) {
            foreach ($purchase_unit['payments']['authorizations'] as $authorization) {
                $authorization_id = $authorization['id'];
                break;
            }
        }

        if (empty($authorization_id)) {
            return false;
        }

        $response = $this->apiCall("v2/payments/authorizations/{$authorization_id}/void", [], 'POST');
        if (empty($response['status_code']) || $response['status_code'] != 204) {
            return false;
        }

        return true;
    }

    public function apiCall($api_point, $in_params = [], $in_http_method = 'GET', $second_call = false)
    {
        global $config;

        $url = $this->getApiEndpoint($api_point);

        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

        if (!empty($in_params['data'])) {
            $http_method = ($in_http_method === 'PATCH' ? 'PATCH' : 'POST');
        } else {
            $http_method = $in_http_method;
        }

        $headers = [
            'Authorization: Bearer ' . $this->getAccessToken(),
            "PayPal-Partner-Attribution-Id: {$config['paypal_checkout_bn_code']}",
            'Content-Type: application/json',
        ];
        if (defined('XC_PAYPAL_CHECKOUT_NEGATIVE_TESTING')) {
            $headers[] = 'PayPal-Mock-Response:' . XC_PAYPAL_CHECKOUT_NEGATIVE_TESTING;
        }
        if (!empty($in_params['headers'])) {
            $headers = array_merge($in_params['headers'], $headers);
        }
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);

        switch ($http_method) {
            case 'POST':
                curl_setopt($ch, CURLOPT_POST, 1);
                break;
            case 'PATCH':
            case 'DELETE':
                curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $http_method);
                break;
        }

        if (in_array($http_method, ['POST', 'PATCH'])) {
            curl_setopt($ch, CURLOPT_POSTFIELDS, empty($in_params['data']) ? '{}' : json_encode($in_params['data']));
        }

        $this->respHeaders = [];
        curl_setopt($ch, CURLOPT_HEADERFUNCTION, [$this, '_writeHeader']);

        $curl_result = curl_exec($ch);
        $curl_info   = curl_getinfo($ch);

        curl_close($ch);

        $curl_result = $curl_result ? json_decode($curl_result, true) : [];

        $this->respHeaders = array_filter($this->respHeaders, function ($header) {
            foreach (['HTTP', 'content-type', 'paypal-debug-id', 'content-length'] as $pattern) {
                if (stripos($header, $pattern) === 0) {
                    return true;
                }
            }

            return false;
        });

        if ("$http_method:$api_point" !== 'GET:v1/notifications/webhooks') {
            $this->logMsg(array_filter([
                'url'             => "$http_method: $url: " . (isset($curl_info['http_code']) ? $curl_info['http_code'] : ''),
                'request_headers' => $headers,
                'request_params'  => !empty($in_params['data']) ? $in_params['data'] : '',
                'response'        => [
                    'headers'     => array_values($this->respHeaders),
                    'result_data' => $curl_result,
                ],
            ]));
        }

        if (
            !empty($curl_result['error'])
            && strpos($curl_result['error'], 'invalid_token') === 0
            && !$second_call
        ) {
            $this->getAccessToken(self::RENEW_ACCESS_TOKEN);

            return $this->apiCall($api_point, $in_params, $in_http_method, true);
        }

        return [
            'result_data' => $curl_result ?: [],
            'status_code' => isset($curl_info['http_code']) ? $curl_info['http_code'] : '',
        ];
    }

    public function getJsonEncode($data)
    {
        return json_encode($data, JSON_UNESCAPED_SLASHES);
    }

    public function logMsg($msg, $force_output = false)
    {
        global $config;

        if (
            $config['Paypal_Checkout']['paypal_checkout_log_requests'] !== 'Y'
            && !$force_output
        ) {
            return;
        }

        if (!is_string($msg)) {
            $msg = preg_replace('/^(  +?)\\1(?=[^ ])/m', '$1', json_encode($msg, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT));
        }

        x_load('logging');
        x_log_add('paypal_checkout', $msg);
    }

    public function check3Dsecure($orderId)
    {
        global $config;

        $result = [
            'passed'  => true,
            'message' => [],
        ];

        $is3dsEnabled = $config['Paypal_Checkout']['paypal_checkout3d_secure'] === 'Y';
        if (!$is3dsEnabled) {
            return $result;
        }

        $order_data = $this->getOrder($orderId);

        $authenticationResult = isset($order_data['result_data']['payment_source']['card']['authentication_result']) ? $order_data['result_data']['payment_source']['card']['authentication_result'] : [];
        $liabilityShift       = isset($authenticationResult['liability_shift']) ? $authenticationResult['liability_shift'] : '';
        $enrollmentStatus     = isset($authenticationResult['three_d_secure']['enrollment_status']) ? $authenticationResult['three_d_secure']['enrollment_status'] : '';
        $authenticationStatus = isset($authenticationResult['three_d_secure']['authentication_status']) ? $authenticationResult['three_d_secure']['authentication_status'] : '';

        if (
            ($enrollmentStatus === 'Y' && $authenticationStatus === 'Y' && $liabilityShift === 'POSSIBLE')
            || ($enrollmentStatus === 'Y' && $authenticationStatus === 'A' && $liabilityShift === 'POSSIBLE')
            || ($enrollmentStatus === 'N' && !$authenticationStatus && $liabilityShift === 'NO')
            || ($enrollmentStatus === 'U' && !$authenticationStatus && $liabilityShift === 'NO')
            || ($enrollmentStatus === 'B' && !$authenticationStatus && $liabilityShift === 'NO')
            || (!$enrollmentStatus && !$authenticationStatus && !$liabilityShift && !$is3dsEnabled)
        ) {
            $result['passed'] = true;
        } elseif (
            ($enrollmentStatus === 'Y' && $authenticationStatus === 'U' && $liabilityShift === 'UNKNOWN')
            || ($enrollmentStatus === 'Y' && $authenticationStatus === 'U' && $liabilityShift === 'NO')
            || ($enrollmentStatus === 'Y' && $authenticationStatus === 'C' && $liabilityShift === 'UNKNOWN')
            || ($enrollmentStatus === 'Y' && !$authenticationStatus && $liabilityShift === 'NO')
            || ($enrollmentStatus === 'U' && !$authenticationStatus && $liabilityShift === 'UNKNOWN')
            || (!$enrollmentStatus && !$authenticationStatus && $liabilityShift === 'UNKNOWN')
        ) {
            $result = [
                'passed'  => false,
                'message' => [
                    'type'    => 'warning',
                    'message' => 'Unable to complete authentication. Please try again.',
                ],
            ];
        } elseif (
            ($enrollmentStatus === 'Y' && $authenticationStatus === 'N' && $liabilityShift === 'NO')
            || ($enrollmentStatus === 'Y' && $authenticationStatus === 'R' && $liabilityShift === 'NO')
        ) {
            $result = [
                'passed'  => false,
                'message' => [
                    'type'    => 'warning',
                    'message' => 'Please, use another card or payment method.',
                ],
            ];
        } else {
            $result = [
                'passed'  => false,
                'message' => [
                    'type'    => 'warning',
                    'message' => 'Unknown error. Please try again.',
                ],
            ];

        }

        if (
            empty($result['passed'])
            && isset($result['message']['message'])
        ) {
            return ['message' => $result['message']['message']];
        }

        return [];
    }

    public function addWebhook($params)
    {
        return $this->apiCall('v1/notifications/webhooks', ['data' => $params], 'POST');
    }

    public function deleteWebhook($webhook_id)
    {
        return $this->apiCall("v1/notifications/webhooks/{$webhook_id}", [], 'DELETE');
    }

    public function getWebhooks()
    {
        return $this->apiCall('v1/notifications/webhooks', [], 'GET')['result_data'];
    }

    private function getOrder($order_id)
    {
        return $this->apiCall("v2/checkout/orders/{$order_id}", [], 'GET');
    }

    private function getAccessToken($force_renew = false)
    {
        global $config, $paypal_checkout_access_token;

        x_session_register('paypal_checkout_access_token', '');

        func_register_cache_function(__FUNCTION__, [
            'class'                => __CLASS__,
            'dir'                  => 'Paypal_Checkout',
            'hashedDirectoryLevel' => 0,
            'ttl'                  => 29160, // default value from paypal 90%
        ]);

        $cache_key = md5("{$config['Paypal_Checkout']['paypal_checkout_client_id']}:{$config['Paypal_Checkout']['paypal_checkout_client_secret']}:{$config['Paypal_Checkout']['pc_merchant_id_of_client']}");

        if ($force_renew === self::RENEW_ACCESS_TOKEN) {
            $paypal_checkout_access_token = '';
        } elseif (
            !$paypal_checkout_access_token
            && ($cacheData = func_get_cache_func($cache_key, __FUNCTION__))
        ) {
            $paypal_checkout_access_token = $cacheData['data'];
        }

        if ($paypal_checkout_access_token) {
            return $paypal_checkout_access_token;
        }

        $url = $this->getApiEndpoint('v1/oauth2/token');

        $ch = curl_init($url);

        $headers = [
            'Accept-Language: en_US',
            'Content-Type: application/json',
        ];

        curl_setopt($ch, CURLOPT_USERPWD, "{$config['Paypal_Checkout']['paypal_checkout_client_id']}:{$config['Paypal_Checkout']['paypal_checkout_client_secret']}");
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_POST, 1);

        curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query([
            'grant_type' => 'client_credentials',
        ]));

        $result = curl_exec($ch);

        curl_close($ch);

        $this->logMsg([
            'url'      => "POST to $url",
            'headers'  => $headers,
            'response' => stripslashes($result),
        ]);

        $result = json_decode($result, true);
        if (empty($result['access_token'])) {
            func_save_cache_func('', '', __FUNCTION__);

            return false;
        }

        $paypal_checkout_access_token = $result['access_token'];
        func_save_cache_func($paypal_checkout_access_token, $cache_key, __FUNCTION__);

        return $paypal_checkout_access_token;
    }

    private function getApiEndpoint($path = '')
    {
        global $config;
        $url = $config['Paypal_Checkout']['paypal_checkout_mode'] === 'test' ? 'https://api-m.sandbox.paypal.com' : 'https://api-m.paypal.com';
        $url .= '/' . ltrim($path, '/');

        return $url;
    }

    private function generateClientToken()
    {
        global $pc_client_token;
        x_session_register('pc_client_token', []);

        if ($pc_client_token && $pc_client_token['expiry'] > XC_TIME) {
            return $pc_client_token['client_token'];
        }

        $response = $this->apiCall('v1/identity/generate-token', [], 'POST');

        if (!empty($response['result_data']['client_token'])) {
            $pc_client_token = [
                'client_token' => $response['result_data']['client_token'],
                'expiry'       => XC_TIME + ($response['result_data']['expires_in'] * 0.80), // 80% time e.g. 50 min
            ];
        } else {
            $pc_client_token = [];
        }

        return $response['result_data']['client_token'];
    }

    private function getPaymentId()
    {
        x_load('payment');
        $tmp = func_get_processor_by_id(XC_PAYPAL_CHECKOUT_CC);

        if ($tmp['paymentid']) {
            return $tmp['paymentid'];
        }

        return 0;
    }

    private function getButtonStyle()
    {
        global $config;
        $result = [];

        foreach (['checkout', 'cart', 'default'] as $section) {
            $layout  = isset($config['Paypal_Checkout']["pc_button_layout_s$section"]) ? $config['Paypal_Checkout']["pc_button_layout_s$section"] : $config["pc_button_layout_s$section"];
            $tagline = !empty($config['Paypal_Checkout']["pc_button_tagline_s$section"]) && $config['Paypal_Checkout']["pc_button_tagline_s$section"] === 'Y';

            if ($layout === 'vertical') {
                $tagline = false; // https://developer.paypal.com/sdk/js/reference/#link-tagline Note: Set the style.layout to horizontal for taglines.
            }

            $result[$section] = json_encode([
                'style' => [
                    'layout'  => $layout,
                    'color'   => $config['Paypal_Checkout']["pc_button_color_s$section"],
                    'shape'   => $config['Paypal_Checkout']["pc_button_shape_s$section"],
                    'label'   => $config['Paypal_Checkout']["pc_button_label_s$section"],
                    'tagline' => $tagline,
                ]]);
        }

        return $result;
    }

    private function getOrderDetails()
    {
        global $cart, $logged_userid, $config, $module_params;

        x_load('user', 'paypal');

        $config_currency = $config['Paypal_Checkout']['paypal_checkout_currency'];
        $userinfo        = func_userinfo($logged_userid, 'C');
        $order_details   = [
            'payer'               => $this->getOrderPayer($userinfo),
            'intent'              => $config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y' ? 'AUTHORIZE' : 'CAPTURE',
            'application_context' => [
                'shipping_preference' => 'SET_PROVIDED_ADDRESS',
            ],
            'purchase_units'      => [
                [
                    'amount' => [
                        'value'         => $cart['total_cost'],
                        'currency_code' => $config_currency,
                    ],
                ],
            ],
        ];

        $module_params = ['params' => ['currency' => $config_currency]];
        $totals        = func_paypal_is_line_items_allowed($cart, $cart['total_cost']);

        if (!empty($totals)) {
            $order_details['purchase_units'][0]['items']               = $this->getOrderItems($cart, $config_currency);
            $order_details['purchase_units'][0]['amount']['breakdown'] = [
                'item_total' => ['currency_code' => $config_currency, 'value' => $totals['ItemTotal']],
                'tax_total'  => ['currency_code' => $config_currency, 'value' => $totals['TaxTotal']],
                'shipping'   => ['currency_code' => $config_currency, 'value' => $totals['ShippingTotal']],
                'handling'   => ['currency_code' => $config_currency, 'value' => $totals['HandlingTotal']],
                'discount'   => ['currency_code' => $config_currency, 'value' => $totals['DiscountTotal']],
            ];
            $order_details['purchase_units'][0]['amount']['breakdown'] = array_filter($order_details['purchase_units'][0]['amount']['breakdown'], function ($item) {
                return !is_null($item['value']);
            });
        }

        if (!func_is_completed_userinfo($userinfo)) {
            unset($order_details['payer']);
            unset($order_details['application_context']);
        }

        if (func_is_completed_userinfo($userinfo)) {
            $order_details['purchase_units'][0]['shipping'] = [
                'type'    => 'SHIPPING',
                'name'    => [
                    'full_name' => trim("{$userinfo['s_firstname']} {$userinfo['s_lastname']}"),
                ],
                'address' => [
                    'address_line_1' => $userinfo['s_address'],
                    'address_line_2' => isset($userinfo['s_address2']) ? $userinfo['s_address2'] : '',
                    'admin_area_1'   => $userinfo['s_state'],
                    'admin_area_2'   => $userinfo['s_city'],
                    'country_code'   => $userinfo['s_country'],
                    'postal_code'    => $userinfo['s_zipcode'],
                ],
            ];
        }

        return $order_details;
    }

    private function _writeHeader($ch, $header)
    {
        if (trim($header)) {
            $this->respHeaders[] = trim($header);
        }

        return strlen($header);
    }

    private function getOrderPayer($userinfo)
    {
        return [
            'email_address' => isset($userinfo['email']) ? $userinfo['email'] : '',
            'address'       => [
                'address_line_1' => isset($userinfo['b_address']) ? $userinfo['b_address'] : '',
                'address_line_2' => isset($userinfo['b_address_2']) ? $userinfo['b_address_2'] : '',
                'admin_area_1'   => isset($userinfo['b_state']) ? $userinfo['b_state'] : '',
                'admin_area_2'   => isset($userinfo['b_city']) ? $userinfo['b_city'] : '',
                'postal_code'    => isset($userinfo['b_zipcode']) ? $userinfo['b_zipcode'] : '',
                'country_code'   => isset($userinfo['b_country']) ? $userinfo['b_country'] : '',
            ],
            'name'          => [
                'given_name' => isset($userinfo['firstname']) ? $userinfo['firstname'] : '',
                'surname'    => isset($userinfo['lastname']) ? $userinfo['lastname'] : '',
            ],
        ];
    }

    private function getOrderItems($cart, $config_currency)
    {
        $items = [];
        if (!empty($cart['products'])) {
            foreach ($cart['products'] as $product) {
                $items[] = [
                    'name'        => substr($product['product'], 0, 127),
                    'unit_amount' => ['currency_code' => $config_currency, 'value' => $product['display_price']],
                    'quantity'    => $product['amount'],
                    'sku'         => substr($product['productcode'], 0, 127),
                    'category'    => 'PHYSICAL_GOODS',
                ];
            }
        }

        return $items;
    }

}
