<?php
/* vim: set ts=4 sw=4 sts=4 et: */
/*****************************************************************************\
 * +-----------------------------------------------------------------------------+
 * | X-Cart Software license agreement                                           |
 * | Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>         |
 * | All rights reserved.                                                        |
 * +-----------------------------------------------------------------------------+
 * | PLEASE READ  THE FULL TEXT OF SOFTWARE LICENSE AGREEMENT IN THE "COPYRIGHT" |
 * | FILE PROVIDED WITH THIS DISTRIBUTION. THE AGREEMENT TEXT IS ALSO AVAILABLE  |
 * | AT THE FOLLOWING URL: https://www.x-cart.com/license-agreement-classic.html |
 * |                                                                             |
 * | THIS AGREEMENT EXPRESSES THE TERMS AND CONDITIONS ON WHICH YOU MAY USE THIS |
 * | SOFTWARE PROGRAM AND ASSOCIATED DOCUMENTATION THAT QUALITEAM SOFTWARE LTD   |
 * | (hereinafter referred to as "THE AUTHOR") OF REPUBLIC OF CYPRUS IS          |
 * | FURNISHING OR MAKING AVAILABLE TO YOU WITH THIS AGREEMENT (COLLECTIVELY,    |
 * | THE "SOFTWARE"). PLEASE REVIEW THE FOLLOWING TERMS AND CONDITIONS OF THIS   |
 * | LICENSE AGREEMENT CAREFULLY BEFORE INSTALLING OR USING THE SOFTWARE. BY     |
 * | INSTALLING, COPYING OR OTHERWISE USING THE SOFTWARE, YOU AND YOUR COMPANY   |
 * | (COLLECTIVELY, "YOU") ARE ACCEPTING AND AGREEING TO THE TERMS OF THIS       |
 * | LICENSE AGREEMENT. IF YOU ARE NOT WILLING TO BE BOUND BY THIS AGREEMENT, DO |
 * | NOT INSTALL OR USE THE SOFTWARE. VARIOUS COPYRIGHTS AND OTHER INTELLECTUAL  |
 * | PROPERTY RIGHTS PROTECT THE SOFTWARE. THIS AGREEMENT IS A LICENSE AGREEMENT |
 * | THAT GIVES YOU LIMITED RIGHTS TO USE THE SOFTWARE AND NOT AN AGREEMENT FOR  |
 * | SALE OR FOR TRANSFER OF TITLE. THE AUTHOR RETAINS ALL RIGHTS NOT EXPRESSLY  |
 * | GRANTED BY THIS AGREEMENT.                                                  |
 * +-----------------------------------------------------------------------------+
 * \*****************************************************************************/

/**
 * PayPal Checkout
 *
 * @category   X-Cart
 * @package    X-Cart
 * @subpackage Modules
 * @author     Ildar Amankulov
 * @copyright  Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>
 * @license    https://www.x-cart.com/license-agreement-classic.html X-Cart license agreement
 * @version    $Id$
 * @link       https://www.x-cart.com/
 * @see        ____file_see____
 */

const PAYPAL_CHECKOUT_SECURE_KEY_LENGTH = 32;

if (!defined('XCART_START')) {
    header('Location: ../../');
    die('Access denied');
}

function func_paypal_checkout_init()
{
    global $xcart_dir;
    require_once "{$xcart_dir}/modules/Paypal_Checkout/class.RestAPI.php";

    if (
        (!defined('AREA_TYPE') || constant('AREA_TYPE') == 'C')
        && !preg_match('%/payment/[^/]*\.php%', $_SERVER['PHP_SELF'])
    ) {
        \XC\PaypalCheckout\RestAPI::getInstance()->assignJsWidgetParams();

        return;
    }

    if (defined('ADMIN_CONFIGURATION_CONTROLLER')) {
        // Register module config update handler
        if (function_exists('func_add_event_listener')) {
            func_add_event_listener('module.config.update', 'func_paypal_checkout_on_module_config_update');
        }
    }

    if (defined('ADMIN_MODULES_CONTROLLER')) {
        // Register module toggle handler
        if (function_exists('func_add_event_listener')) {
            func_add_event_listener('module.ajax.toggle', 'func_paypal_checkout_on_module_toggle');
        }
    }
}

/**
 * @throws Exception
 */
function func_paypal_checkout_do_capture($order)
{
    if (!empty($order['extra']['paypal_checkout_capture'])) {
        return true;
    }

    $paypal_checkout_order_data = json_decode($order['extra']['paypal_checkout_order'], true);

    if (empty($paypal_checkout_order_data)) {
        throw new Exception(__FUNCTION__ . ': PayPal order is missing');
    }

    $captured_response = \XC\PaypalCheckout\RestAPI::getInstance()->captureAuthorizedPayment($paypal_checkout_order_data, $order['orderid']);

    if (false === $captured_response) {
        throw new Exception(__FUNCTION__ . ': Order was not captured');
    }

    $response = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/captures/{$captured_response['id']}", [], 'GET');
    if ($response['status_code'] == 200) {
        func_array2insert('order_extras', [
            'orderid' => $order['orderid'],
            'khash'   => 'paypal_checkout_capture',
            'value'   => addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($response['result_data'])),
        ], true);
    }

    return true;
}

/**
 * @throws Exception
 */
function func_paypal_checkout_do_refund($order, $amount = false)
{
    global $config;

    if (!empty($order['extra']['paypal_checkout_refund'])) {
        return true;
    }

    $paypal_capture_data = json_decode($order['extra']['paypal_checkout_capture'], true);

    if (empty($paypal_capture_data)) {
        throw new Exception(__FUNCTION__ . ': PayPal order is missing');
    }

    if (false !== $amount) {
        $amount = [
            'currency_code' => $config['Paypal_Checkout']['paypal_checkout_currency'],
            'value'         => floatval($amount),
        ];
    }

    $refunded_response = \XC\PaypalCheckout\RestAPI::getInstance()->refundOrder($paypal_capture_data, $order['orderid'], $amount);

    if (false === $refunded_response) {
        throw new Exception(__FUNCTION__ . ': PayPal order was not returning');
    }

    $response = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/refunds/{$refunded_response['id']}", [], 'GET');

    if ($response['status_code'] == 200) {
        func_array2insert('order_extras', [
            'orderid' => $order['orderid'],
            'khash'   => 'paypal_checkout_refund',
            'value'   => addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($response['result_data'])),
        ], true);
    }

    return true;
}

/**
 * @throws Exception
 */
function func_paypal_checkout_do_void($order)
{
    if (!empty($order['extra']['paypal_checkout_void'])) {
        return true;
    }

    $paypal_checkout_order_data = json_decode($order['extra']['paypal_checkout_order'], true);
    if (empty($paypal_checkout_order_data)) {
        throw new Exception(__FUNCTION__ . ':PayPal order is missing');
    }

    $result = \XC\PaypalCheckout\RestAPI::getInstance()->voidOrder($paypal_checkout_order_data);
    if (false === $result) {
        throw new Exception(__FUNCTION__ . ':Payment was not voided');
    }

    func_array2insert('order_extras', [
        'orderid' => $order['orderid'],
        'khash'   => 'paypal_checkout_void',
        'value'   => 1,
    ], true);

    return true;
}

function func_paypal_checkout_filter_config($options)
{
    $settings_form_fields    = [];
    $customized_form_fields  = [];
    $credit_messaging_fields = [];

    foreach ($options as $k => $v) {
        if (
            $v['type'] === 'selector'
            || $v['type'] === 'multiselector'
        ) {
            $vars = func_parse_str(trim($v['variants']), "\n", ":");
            $vars = func_array_map('trim', $vars);

            // Check variable data
            if ($v['type'] == "multiselector") {
                $options[$k]['value'] = $v['value'] = explode(";", $v['value']);

                foreach ($v['value'] as $vk => $vv) {
                    if (!isset($vars[$vv]))
                        unset($v['value'][$vk]);
                }

                $options[$k]['value'] = $v['value'] = array_values($v['value']);
            }

            $options[$k]['variants'] = [];

            foreach ($vars as $vk => $vv) {
                if (!empty($vv) && strpos($vv, "_") !== false && strpos($vv, " ") === false) {
                    if (!empty($name)) {
                        $vv = $name;
                    }
                }

                $options[$k]['variants'][$vk] = ['name' => $vv];
            }

            foreach ($options[$k]['variants'] as $vk => $vv) {
                $options[$k]['variants'][$vk]['selected'] = $options[$k]['type'] === 'selector'
                    ? $options[$k]['value'] == $vk
                    : in_array($vk, $options[$k]['value']);

            }
        }
    }

    foreach ($options as $option) {
        if (stripos($option['name'], 'pc_button_') === 0) {
            $customized_form_fields[] = $option;
        } elseif (stripos($option['name'], 'pc_credit_') === 0) {
            $credit_messaging_fields[] = $option;
        } else {
            $settings_form_fields[] = $option;
        }
    }

    return [$settings_form_fields, $customized_form_fields, $credit_messaging_fields];
}

function func_paypal_checkout_get_ccprocessor_data($integration_type)
{
    global $config;
    $result = [];

    // Prepare data to insert to DB
    $db_data = [
        XC_PAYPAL_CHECKOUT_CC => [
            'module_name'       => 'PayPal Checkout',
            'type'              => 'P', // Payment systems
            'processor'         => XC_PAYPAL_CHECKOUT_CC,
            'template'          => 'cc_paypal_checkout.tpl',
            'param02'           => $config['Paypal_Checkout']['paypal_checkout_client_id'],
            'param03'           => $config['Paypal_Checkout']['paypal_checkout_client_secret'],
            'param04'           => $config['Paypal_Checkout']['pc_merchant_id_of_client'],
            'param05'           => $config['Paypal_Checkout']['paypal_checkout_currency'],
            'param06'           => '', // prefix
            'disable_ccinfo'    => 'Y',
            'background'        => 'N',
            'testmode'          => $config['Paypal_Checkout']['paypal_checkout_mode'] === 'test' ? 'Y' : 'N',
            'is_check'          => '',
            'is_refund'         => '',
            'c_template'        => 'modules/Paypal_Checkout/payment.tpl',
            'cmpi'              => '',
            'use_preauth'       => $config['Paypal_Checkout']['paypal_checkout_use_preauth'],
            'preauth_expire'    => '0',
            'has_preauth'       => '',
            'capture_min_limit' => '0%',
            'capture_max_limit' => '0%',
        ],
    ];

    if (isset($db_data[$integration_type])) {
        $result = $db_data[$integration_type];
    }

    return $result;
}

function func_paypal_checkout_on_module_enable()
{
    x_load('payment');

    // Add processors data to DB
    $data_from_module_config = func_paypal_checkout_get_ccprocessor_data(XC_PAYPAL_CHECKOUT_CC);
    func_array2insert('ccprocessors', $data_from_module_config);

    func_paypal_checkout_register_webhooks([
        'paypal_checkout_client_id'     => $data_from_module_config['param02'],
        'paypal_checkout_client_secret' => $data_from_module_config['param03'],
    ]);

    // Register processors
    func_add_processor(XC_PAYPAL_CHECKOUT_CC);
}

function func_paypal_checkout_on_module_disable()
{
    global $sql_tbl;

    x_load('payment');

    // Unregister processor
    func_remove_processor(XC_PAYPAL_CHECKOUT_CC);

    // must be before delete
    func_paypal_checkout_delete_webhooks();

    // Remove processor from DB
    db_query("DELETE FROM $sql_tbl[ccprocessors] WHERE processor='" . XC_PAYPAL_CHECKOUT_CC . "'");
}

function func_paypal_checkout_on_module_toggle($module_name, $module_new_state, $previous_handler_result = '')
{
    $redirect = $previous_handler_result;

    if ($module_name === XC_PAYPAL_CHECKOUT_MODULE_NAME) {
        switch ($module_new_state) {
            case true:
                func_paypal_checkout_on_module_enable();
                break;
            default:
                func_paypal_checkout_on_module_disable();
                break;
        }
    }

    return $redirect;
}

function func_paypal_checkout_on_module_config_update($option, $section_data)
{
    if (!empty($section_data['paypal_checkout_mode'])) {
        // Update processor data in DB
        func_array2update(
            'ccprocessors',
            [
                'param02'     => $section_data['paypal_checkout_client_id'],
                'param03'     => $section_data['paypal_checkout_client_secret'],
                'param04'     => $section_data['pc_merchant_id_of_client'],
                'param05'     => $section_data['paypal_checkout_currency'],
                'use_preauth' => $section_data['paypal_checkout_use_preauth'],
                'testmode'    => $section_data['paypal_checkout_mode'] === 'test' ? 'Y' : 'N',
            ],
            "processor='" . XC_PAYPAL_CHECKOUT_CC . "'"
        );

        func_paypal_checkout_register_webhooks($section_data);
    }
}

function func_paypal_checkout_register_webhooks($config_data)
{
    global $https_location, $top_message, $config, $sql_tbl;
    if (
        empty($config_data['paypal_checkout_client_id'])
        || empty($config_data['paypal_checkout_client_secret'])
        || stripos($https_location, '/xlocal/') !== false
        || stripos($https_location, '/localhost/') !== false
        || $config['paypal_checkout_enable_webhooks'] !== 'Y'
    ) {
        return;
    }
    $cache_name = func_paypal_checkout_cache_webhook_state();
    $cache_key  = md5("$https_location:{$config_data['paypal_checkout_client_id']}:{$config_data['paypal_checkout_client_secret']}");
    if (
        ($cacheData = func_get_cache_func($cache_key, $cache_name))
        && $cacheData['data']
        && !empty($config['paypal_checkout_secure_key'])
    ) {
        return;
    }

    $config['Paypal_Checkout']['paypal_checkout_client_id']     = $config_data['paypal_checkout_client_id'];
    $config['Paypal_Checkout']['paypal_checkout_client_secret'] = $config_data['paypal_checkout_client_secret'];

    $is_ok = func_paypal_checkout_delete_webhooks();
    if (!$is_ok) {
        // cannot remove previous webhooks
        func_save_cache_func(false, $cache_key, $cache_name);

        return;
    }

    $secureKey   = func_get_secure_random_key(PAYPAL_CHECKOUT_SECURE_KEY_LENGTH, 'user_password');
    $callbackUrl = "$https_location/payment/cc_paypal_checkout.php?mode=webhook&key=$secureKey";
    // https://developer.paypal.com/docs/api/webhooks/v1/#event-type_list
    $params = [
        'url'         => $callbackUrl,
        'event_types' => [
            ['name' => 'PAYMENT.AUTHORIZATION.VOIDED'],
            ['name' => 'PAYMENT.CAPTURE.COMPLETED'],
            ['name' => 'PAYMENT.CAPTURE.DENIED'],
            //['name' => 'PAYMENT.CAPTURE.REFUNDED'], currently is not supported
        ],
    ];

    $response = \XC\PaypalCheckout\RestAPI::getInstance()->addWebhook($params);
    if (!empty($response['status_code']) && $response['status_code'] == 201) {
        $secureKey = addslashes($secureKey);
        db_query("UPDATE $sql_tbl[config] SET `value` = '$secureKey' WHERE `name` = 'paypal_checkout_secure_key'");
        func_save_cache_func(true, $cache_key, $cache_name);
    } else {
        if (isset($response['result_data']['details'][0]['issue']) ? $response['result_data']['details'][0]['issue'] : '') {
            if ($top_message['content']) {
                $top_message = ['content' => ''];
            }
            $top_message['content'] .= "<br />" . $response['result_data']['details'][0]['issue'] . " $callbackUrl";
            $top_message['type']    = 'W';
        }
        func_save_cache_func(false, $cache_key, $cache_name);
    }
}

function func_paypal_checkout_delete_webhooks()
{
    global $config, $https_location, $top_message;
    if (
        empty($config['Paypal_Checkout']['paypal_checkout_client_id'])
        || empty($config['Paypal_Checkout']['paypal_checkout_client_secret'])
        || $config['paypal_checkout_enable_webhooks'] !== 'Y'
    ) {
        return true;
    }

    $response = \XC\PaypalCheckout\RestAPI::getInstance()->getWebhooks();
    if (!empty($response['webhooks'])) {
        $cache_name = func_paypal_checkout_cache_webhook_state();
        $cache_key  = md5("$https_location:{$config['Paypal_Checkout']['paypal_checkout_client_id']}:{$config['Paypal_Checkout']['paypal_checkout_client_secret']}");

        foreach ($response['webhooks'] as $webhook) {
            if (stripos($webhook['url'], $https_location) !== false) {
                func_save_cache_func(false, $cache_key, $cache_name);
                \XC\PaypalCheckout\RestAPI::getInstance()->deleteWebhook($webhook['id']);
            } else {
                $is_failed = false;
                if ($top_message['content']) {
                    $top_message = ['content' => ''];
                }
                $top_message['content'] .= "{$webhook['url']}<br /> webhook cannot be deleted as it doesn't match your current shop URL<br />$https_location<br />";
                $top_message['type']    = 'W';
                \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("{$webhook['url']} webhook cannot be deleted as it doesn't match your current shop URL $https_location", true);
            }
        }
    }

    return isset($is_failed) ? $is_failed : true;
}

function func_paypal_checkout_request_webhooks($configuration)
{
    global $config, $sql_tbl, $https_location;
    if ($config['paypal_checkout_enable_webhooks'] !== 'Y') {
        return $configuration;
    }

    $url = $old_value = 'NONE';
    if (
        !empty($config['Paypal_Checkout']['paypal_checkout_client_id'])
        && !empty($config['Paypal_Checkout']['paypal_checkout_client_secret'])
    ) {
        $response = \XC\PaypalCheckout\RestAPI::getInstance()->getWebhooks();
        if (!empty($response['webhooks'])) {
            $urls = array_reduce($response['webhooks'], function ($carry, $item) {
                return array_merge($carry, [$item['url']]);
            }, []) ?: [];
            $url  = implode('<br />', $urls) ?: 'NONE';
        }
    }

    $result_value = "Webhook URL assigned to the current Client ID/Secret:<br /><div id=\"paypal-checkout-webhookurl-comment\" style=\"color: #2c3e49;\">{$url}</div>";
    if ($url !== 'NONE') {
        if (stripos($url, $https_location) === false) {
            $result_value .= "<div style=\"color: red;\">It doesn't match your current shop domain $https_location</div>";
        }
        if (!$config['paypal_checkout_secure_key']) {
            $result_value .= "<div style=\"color: red;\">Your current secure key is empty.</div>";
        } elseif (strpos($url, $config['paypal_checkout_secure_key']) === false) {
            $result_value .= "<div style=\"color: red;\">Key from the URL doesn't match your current secure key: {$config['paypal_checkout_secure_key']}</div>";
        }
    }

    foreach ($configuration as $k => $v) {
        if ($v['name'] != 'pc_webhook_url_comment') {
            continue;
        }

        $old_value                    = $v['value'];
        $configuration[$k]['comment'] = $result_value;
        break;
    }
    if ($old_value !== $url) {
        // For logging
        $url = addslashes($url);
        db_query("UPDATE $sql_tbl[config] SET value='$url' WHERE `name`='pc_webhook_url_comment'");
    }

    return $configuration;
}

/**
 * Keeps state of webhooks
 */
function func_paypal_checkout_cache_webhook_state()
{
    func_register_cache_function(__FUNCTION__, [
        'dir'                  => 'Paypal_Checkout',
        'hashedDirectoryLevel' => 0,
        'ttl'                  => 86400 * 30,
    ]);

    return __FUNCTION__;
}
