<?php
/* vim: set ts=4 sw=4 sts=4 et: */
/*****************************************************************************\
 * +-----------------------------------------------------------------------------+
 * | X-Cart Software license agreement                                           |
 * | Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>         |
 * | All rights reserved.                                                        |
 * +-----------------------------------------------------------------------------+
 * | PLEASE READ  THE FULL TEXT OF SOFTWARE LICENSE AGREEMENT IN THE "COPYRIGHT" |
 * | FILE PROVIDED WITH THIS DISTRIBUTION. THE AGREEMENT TEXT IS ALSO AVAILABLE  |
 * | AT THE FOLLOWING URL: https://www.x-cart.com/license-agreement-classic.html |
 * |                                                                             |
 * | THIS AGREEMENT EXPRESSES THE TERMS AND CONDITIONS ON WHICH YOU MAY USE THIS |
 * | SOFTWARE PROGRAM AND ASSOCIATED DOCUMENTATION THAT QUALITEAM SOFTWARE LTD   |
 * | (hereinafter referred to as "THE AUTHOR") OF REPUBLIC OF CYPRUS IS          |
 * | FURNISHING OR MAKING AVAILABLE TO YOU WITH THIS AGREEMENT (COLLECTIVELY,    |
 * | THE "SOFTWARE"). PLEASE REVIEW THE FOLLOWING TERMS AND CONDITIONS OF THIS   |
 * | LICENSE AGREEMENT CAREFULLY BEFORE INSTALLING OR USING THE SOFTWARE. BY     |
 * | INSTALLING, COPYING OR OTHERWISE USING THE SOFTWARE, YOU AND YOUR COMPANY   |
 * | (COLLECTIVELY, "YOU") ARE ACCEPTING AND AGREEING TO THE TERMS OF THIS       |
 * | LICENSE AGREEMENT. IF YOU ARE NOT WILLING TO BE BOUND BY THIS AGREEMENT, DO |
 * | NOT INSTALL OR USE THE SOFTWARE. VARIOUS COPYRIGHTS AND OTHER INTELLECTUAL  |
 * | PROPERTY RIGHTS PROTECT THE SOFTWARE. THIS AGREEMENT IS A LICENSE AGREEMENT |
 * | THAT GIVES YOU LIMITED RIGHTS TO USE THE SOFTWARE AND NOT AN AGREEMENT FOR  |
 * | SALE OR FOR TRANSFER OF TITLE. THE AUTHOR RETAINS ALL RIGHTS NOT EXPRESSLY  |
 * | GRANTED BY THIS AGREEMENT.                                                  |
 * +-----------------------------------------------------------------------------+
 * \*****************************************************************************/

/**
 * PayPal Checkout
 *
 * @category   X-Cart
 * @package    X-Cart
 * @subpackage Modules
 * @author     Ildar Amankulov
 * @copyright  Copyright (c) 2001-present Qualiteam software Ltd <info@x-cart.com>
 * @license    https://www.x-cart.com/license-agreement-classic.html X-Cart license agreement
 * @version    $Id$
 * @link       https://www.x-cart.com/
 * @see        ____file_see____
 */

// commit=false scenario, step 1
if ($_SERVER['REQUEST_METHOD'] === 'GET' && $_GET['mode'] === 'express_return' && !empty($_GET['token'])) {
    // return from PayPal after step 1, Express started
    if (!defined('XCART_START')) {
        require_once __DIR__ . '/auth.php';
    }

    x_load('payment'); // for fUnc_caRt_gET_paYpAl_expireSs_id and fUnc_Pm_loAd
    $paymentid = isset($paymentid) ? $paymentid : 0;
    func_pm_load('cc_paypal_checkout');
    func_pc_handler2_update_xc_customer_data($_GET['token'],
        isset($pp_subject) ? $pp_subject : '',
        isset($useraction) ? $useraction : '',
        isset($RefererCookie) ? $RefererCookie : ''
    );
    exit;
}
/**
 * @var string $REQUEST_METHOD
 * @var string $xcart_dir
 * @var string $call_type
 * @var string $XCARTSESSID
 * @var array  $secure_oid
 * @var array  $config
 * @var array  $sql_tbl
 */

if (!empty($_REQUEST['call_type']) && in_array($_REQUEST['call_type'], ['call_create_order', 'start_express_checkout', 'call_authorize_order'])) {
    if (!defined('XCART_START')) {
        require __DIR__ . '/auth.php';
    }
    x_load('cart', 'payment'); // for fUnc_caRt_gET_paYpAl_expireSs_id and fUnc_Pm_loAd
    if (in_array($call_type, ['call_create_order', 'start_express_checkout'])) { // start express checkout
        func_pm_load('cc_paypal_checkout');
        echo func_pc_handler1_create_order(isset($paymentid) ? $paymentid : func_cart_get_paypal_express_id(), $call_type);
    }

    if ($call_type === 'call_authorize_order') {
        // https://developer.paypal.com/docs/multiparty/checkout/save-payment-methods/during-purchase/js-sdk/cards/#link-authorizeorcaptureorderandsavecard
        header('Content-type: application/json');
        /** @var string $orderId */
        $pc_3d_secure_failed = empty($is_paypal_method) ? \XC\PaypalCheckout\RestAPI::getInstance()->check3Dsecure($orderId) : [];
        if (!empty($pc_3d_secure_failed['message'])) {
            echo json_encode(['errorMsg' => $pc_3d_secure_failed['message']], JSON_FORCE_OBJECT);
        } else {
            if ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y') {
                echo json_encode(\XC\PaypalCheckout\RestAPI::getInstance()->authorizeOrder($orderId));
            } else {
                echo json_encode(\XC\PaypalCheckout\RestAPI::getInstance()->captureOrder($orderId));
            }
        }
    }

    exit;
}

// webhook calls from PayPal
if ($_SERVER['REQUEST_METHOD'] === 'POST' && !empty($_GET['mode']) && $_GET['mode'] === 'webhook') {
    if (!defined('XCART_START')) {
        require_once __DIR__ . '/auth.php';
    }

    if ($config['paypal_checkout_enable_webhooks'] && $config['paypal_checkout_enable_webhooks'] !== 'Y') {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("WEBHOOK: Callback cannot be processed as it disabled via paypal_checkout_enable_webhooks", true);
        exit;
    }

    if (
        empty($_GET['key'])
        || $_GET['key'] !== $config['paypal_checkout_secure_key']
    ) {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("WEBHOOK: Callback cannot be processed as the keys don't match. Paypal key:'{$_GET['key']}', Your key:'{$config['paypal_checkout_secure_key']}' ", true);
        exit;
    }
    $data = @file_get_contents('php://input');

    if (
        !$data
        || !($data = json_decode($data, true))
        || empty($data['event_type'])
    ) {
        \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("WEBHOOK: Callback cannot be processed as the data is empty or cannot be decoded", true);
        exit;
    }
    \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("WEBHOOK: from PayPal {$data['event_type']}:" . print_r($data, true));

    switch ($data['event_type']) {
        case 'PAYMENT.CAPTURE.COMPLETED':
        case 'PAYMENT.CAPTURE.DENIED':
        case 'PAYMENT.AUTHORIZATION.VOIDED':
            if ($data['event_type'] === 'PAYMENT.AUTHORIZATION.VOIDED') {
                $_entity_response    = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/authorizations/{$data['resource']['id']}", [], 'GET') ?: [];
                $bill_output['code'] = 2;
            } else {
                $_entity_response    = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/captures/{$data['resource']['id']}", [], 'GET') ?: [];
                $bill_output['code'] = $data['resource']['status'] === 'COMPLETED' ? 1 : 2;
            }
            if (
                empty($_entity_response['result_data']['status'])
                || $_entity_response['result_data']['status'] !== $data['resource']['status']
            ) {
                // security failed
                exit;
            }

            $bill_output['billmes'] = " ({$data['event_type']}, Status:{$data['resource']['status']}) ";
            $bill_output['billmes'] .= " (Transaction ID: {$data['resource']['id']}) ";
            $bill_output['billmes'] .= " ({$data['summary']}) ";

            $skey = isset($data['resource']['supplementary_data']['related_ids']['order_id']) ? $data['resource']['supplementary_data']['related_ids']['order_id'] : '';

            require $xcart_dir . '/payment/payment_ccmid.php';
            require $xcart_dir . '/payment/payment_ccwebset.php';
            break;

        default;
            \XC\PaypalCheckout\RestAPI::getInstance()->logMsg("WEBHOOK: Handling for '{$data['event_type']}' is not supported. (Transaction ID: {$data['resource']['id']}).\n{$data['summary']}", true);
            exit;
    }

    exit; // 200 OK here
}

#
# Inclusions from payment/payment_cc.php
#
if (!defined('XCART_START')) {
    header("Location: ../");
    die("Access denied");
}

// commit=false scenario(double step), step 2
if ($REQUEST_METHOD == 'POST' && $_POST['action'] == 'place_order' && empty($_POST['paypal_checkout_order_id'])) {
    // finish ExpressCheckout
    func_pc_handler3_update_n_finish_order();
    require $xcart_dir . '/payment/payment_ccend.php';
    exit;
}

// commit=true scenario(hosted cards) from checkout page
try {
    if (empty($_POST['paypal_checkout_order_id'])) {
        throw new Exception("empty(\$paypal_checkout_order_id):{$_POST['paypal_checkout_order_id']}");
    }

    $paypal_checkout_order_id = $_POST['paypal_checkout_order_id'];

    if (func_query_first_cell("SELECT COUNT(*) FROM $sql_tbl[order_extras] WHERE khash = 'paypal_checkout_id' AND value = '{$paypal_checkout_order_id}'")) {
        throw new Exception("\$paypal_checkout_order_id cannot be used:{$paypal_checkout_order_id}");
    }

    $response = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/checkout/orders/{$paypal_checkout_order_id}", [], 'GET');

    $paypal_checkout_order_data = isset($response['result_data']) ? $response['result_data'] : '';

    if (
        empty($paypal_checkout_order_data)
        || $response['status_code'] != 200
    ) {
        throw new Exception('Order was not found on paypal side:' . print_r($paypal_checkout_order_data, true));
    }
    $extra_order_data = [];

    if ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y') {
        $extra_order_data['capture_status'] = 'A';
    } else {
        $extra_order_data['capture_status'] = '';

        $_captured_data = isset($paypal_checkout_order_data['purchase_units'][0]['payments']['captures'][0]) ? $paypal_checkout_order_data['purchase_units'][0]['payments']['captures'][0] : [];

        if (!$_captured_data) {
            throw new Exception("Capture cannot be done:{$paypal_checkout_order_data['id']}");
        }

        $response = \XC\PaypalCheckout\RestAPI::getInstance()->apiCall("v2/payments/captures/{$_captured_data['id']}", [], 'GET');
        if ($response['status_code'] == 200) {
            $extra_order_data['paypal_checkout_capture'] = addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($response['result_data']));
        }
    }

    $extra_order_data['paypal_checkout_id']    = addslashes($paypal_checkout_order_data['id']);
    $extra_order_data['paypal_checkout_order'] = addslashes(\XC\PaypalCheckout\RestAPI::getInstance()->getJsonEncode($paypal_checkout_order_data));
    if (version_compare($config['version'], '4.5.0', '>=')) {
        db_query("REPLACE INTO $sql_tbl[cc_pp3_data] (`ref`,`sessid`,`trstat`) VALUES ('{$extra_order_data['paypal_checkout_id']}','$XCARTSESSID','GO|" . implode('|', $secure_oid) . "')");
    } else {
        db_query("REPLACE INTO $sql_tbl[cc_pp3_data] (`ref`,`sessionid`,`trstat`) VALUES ('{$extra_order_data['paypal_checkout_id']}','$XCARTSESSID','GO|" . implode('|', $secure_oid) . "')");
    }

    $bill_output = [];
    if ($paypal_checkout_order_data['status'] != 'COMPLETED') {
        throw new Exception("Payment status:{$paypal_checkout_order_data['status']}");
    } else {
        $bill_output['code'] = 1;

        if ($config['Paypal_Checkout']['paypal_checkout_use_preauth'] === 'Y') {
            $bill_output['is_preauth'] = true;
        }

        $bill_output['billmes'] = "Payment complete (PayPal Order ID: {$paypal_checkout_order_data['id']}) ";
    }
} catch (Exception $e) {
    $bill_output['billmes'] = $e->getMessage();;
    $bill_output['code'] = 2;
}
