/* vim: set ts=2 sw=2 sts=2 et: */

if (typeof window.XCPaypalGeneralCheckout !== 'function') {
  window.XCPaypalGeneralCheckout = function (event) {
    if (event) {
      event.stopPropagation();
    }
    let btnConfig;

    $(document).ready(function () {
      btnConfig = window.paypalCheckoutConfig;
    });

    $(document).on('click', '#paypal-checkout-cc-fields-submit', function (e) {
      e.stopPropagation();
      e.preventDefault();

      // check checkboxes
      const checkboxes2check = [{selector: '#accept_terms', msg: txt_accept_terms_err}];
      if (typeof txt_gdpr_accept_terms_err !== 'undefined') {//if $active_modules.EU_Cookie_Law EU_GDPR
        checkboxes2check.push({selector: '#accept_gdpr_checkout', msg: txt_gdpr_accept_terms_err});
      }
      const has_shake_effect = $.effects && $.effects.effect && $.effects.effect.shake;
      for (let index = 0; index < checkboxes2check.length; ++index) {
        const termsObj = $(checkboxes2check[index].selector);
        if (termsObj.length && !termsObj[0].checked) {
          if (has_shake_effect) {
            $('button.place-order-button').effect('shake', {'distance': 4, times: 2});// if $active_modules.EU_Cookie_Law EU_GDPR
          } else {
            xAlert(checkboxes2check[index].msg, '', 'W');
          }
          termsObj.css('outline', 'red solid thin');
          termsObj.one('click', function () {
            $(this).css('outline', '');
          });
          return false;
        }
      } // for (var index = 0; index < checkboxes2check.length; ++index)

      // https://developer.paypal.com/docs/checkout/advanced/integrate/#link-integratefrontend
      let params = {};
      if (document.getElementById('paypal-checkout-customer')) {
        (obj => {
          params = {
            'billingAddress': {
              'streetAddress': obj.b_address,
              'extendedAddress': obj.b_address_2,
              'region': obj.b_state,
              'locality': obj.b_city,
              'postalCode': obj.b_zipcode,
              'countryCodeAlpha2': obj.b_country
            },
            'cardholderName': obj.bl_firstname
          };
        })(document.getElementById('paypal-checkout-customer').dataset);
      }

      if (btnConfig['paypal_checkout_3d_secure']) {
        // https://developer.paypal.com/docs/checkout/advanced/customize/3d-secure/sdk/#link-updatetheadvancedcardfieldscode
        params.contingencies = [btnConfig['paypal_checkout_3d_secure']];
      }

      hostedFieldsHandler.submit(params).then(function (orderData) {
        if (
          orderData.orderId
          && (
            !orderData.liabilityShift
            || orderData.liabilityShift === 'POSSIBLE'
          )
        ) {
          showXCblockUI();
          return $.post('payment/cc_paypal_checkout.php', {
            orderId: orderData.orderId,
            call_type: 'call_authorize_order'
          }).then(function (orderData) {
            if (orderData['errorMsg']) {
              throw new Error(orderData['errorMsg']);
            }
            $('#paypal-checkout-order-id').val(orderData.id);
            $(document).trigger('event-paypal-onApprove-end');
          });
        } else {
          throw new Error(JSON.stringify(orderData));
        }
      }).catch(function (error) {
        $.unblockUI();
        showTopMessage($('#paypal-checkout-customer').data('txt_paypal_checkout_unprocessable_entity') || 'Your payment could not be processed at this time. Please make sure the card information was entered correctly and resubmit. If the problem persists, please contact your credit card company to authorize the purchase', 'e');
        console.log(error);
      });

      return false;
    });

    let hostedFieldsHandler;

    window.PayPalCheckout = {
      pcRenderPayPalButtons: function () {
        let buttonsConfig = Object.assign(paypalCheckoutButtonConfig, {
          createOrder: function (data, actions) {
            return $.post('payment/cc_paypal_checkout.php', {
              call_type: 'call_create_order'
            }).then(function (response) {
              return response;
            });
          },

          onError: function (err) {
          },
          onCancel: function (data) {
          },
          onApprove: function (data, actions) {
            showXCblockUI();
            return $.post('payment/cc_paypal_checkout.php', {
              orderId: data.orderID,
              call_type: 'call_authorize_order',
              is_paypal_method: 'true'
            }).then(function (response) {
              return response;
            }).then(function (orderData) {
              $('#paypal-checkout-order-id').val(orderData.id);
              $(document).trigger('event-paypal-onApprove-end');
            });
          }
        });

        if (!btnConfig.is_checkout_page || !btnConfig.is_flc) {
          paypal.Buttons(buttonsConfig).render('#pcb-container');
        }

        if (btnConfig.is_checkout_page && paypal.HostedFields.isEligible()) {

          if (document.getElementById('paypal-checkout-cc-fields')) {
            $('#pcb-container').after(document.getElementById('paypal-checkout-cc-fields').innerHTML);
          }

          paypal.HostedFields.render({
            createOrder: function () {
              return $.post('payment/cc_paypal_checkout.php', {
                call_type: 'call_create_order'
              }).then(function (response) {
                return response;
              });
            },
            styles: { // https://developer.paypal.com/docs/checkout/advanced/integrate/#link-integratefrontend
              '.valid': {
                color: 'green'
              },
              '.invalid': {
                color: 'red'
              }
            },
            fields: {
              number: {
                selector: '#pcp-hosted-fields-card-number',
                placeholder: 'Credit Card Number'
              },
              cvv: {
                selector: '#pcp-hosted-fields-cvv',
                placeholder: 'CVV'
              },
              expirationDate: {
                selector: '#pcp-hosted-fields-expiration-date',
                placeholder: 'MM/YYYY'
              }
            }
          }).then(function (hostedFieldsInstance) {
            hostedFieldsHandler = hostedFieldsInstance;
          });
        }
      }
    };

    $(document).ready(function () {
      if (typeof ajax.widgets.checkout !== 'undefined') {
        // OPC
        $(document).ready(function () {
          payment_id = parseInt(window.paypalCheckoutPaymentId);
          opcCheckoutWidget = document.querySelector('.opc-container').checkoutWidget;
          $(document).trigger('event-opc-PaymentChangedCustom');
        });

        $(document).on('event-paypal-onApprove-end', function () {
          $('form[name=checkout_form]', opcCheckoutWidget.elm$).trigger('submit');
        });

        $(document).on('event-opc-PaymentChanged event-opc-PaymentChangedCustom', function () {
          $('#paypal-checkout').remove();

          if (payment_id != opcCheckoutWidget.paymentid) {
            $(document).off('input.paypalCheckout');
            $('#btn_box').show();
          } else {
            $('#btn_box').after('<div class="paypal-checkout" id="paypal-checkout" ><div id="pcb-container"></div></div>');
            $('#btn_box').hide();
            PayPalCheckout.pcRenderPayPalButtons();
            hideShowButtons();
            $(document).on('input.paypalCheckout', 'form[name=checkout_form] input:checkbox', hideShowButtons);
          }
        });

        let opcCheckoutWidget, payment_id;

        function hideShowButtons (event_type) {
          if (
            !checkCheckoutForm(document.querySelector('form[name=checkout_form]'))
            || !checkFormFields($('form[name=paymentform]', opcCheckoutWidget.payment).get(0))
            || !opcCheckoutWidget.checkElement()
            || !opcCheckoutWidget.isReady()
          ) {
            $('#paypal-checkout').addClass('paypal-checkout_inactive');
          } else {
            $('#paypal-checkout').removeClass('paypal-checkout_inactive');
          }
        }
      } else {
        // FLC
        $(document).on('event-paypal-onApprove-end', function () {
          $('#paypal-checkout').remove();
          $('#btn_box').show();

          $('form[name=checkout_form]').trigger('submit');
        });

        $(document).ready(function () {
          if (!document.getElementById('paypal-checkout-order-id')) {
            return;
          }
          $('#paypal-checkout').remove();

          $('#btn_box').hide();
          $('#btn_box').after('<div class="paypal-checkout" id="paypal-checkout" ><div id="pcb-container"></div></div>');

          PayPalCheckout.pcRenderPayPalButtons();
        });
      }
    });
  };
}

if (typeof paypal !== 'undefined' && typeof paypal.Buttons().render !== 'undefined') {
  XCPaypalGeneralCheckout();
} else {
  document.addEventListener('PayPal_lib_checkout_is_loaded', XCPaypalGeneralCheckout);
}
